package org.netkernel.lang.dpml.ast.impl;

import org.netkernel.container.ILogger;
import org.netkernel.lang.dpml.ast.*;
import org.netkernel.lang.dpml.endpoint.DPMLRuntime;
import org.netkernel.lang.dpml.endpoint.DPMLScriptContext;
import org.netkernel.lang.dpml.endpoint.DPMLSubrequestContext;
import org.netkernel.layer0.nkf.INKFRequest;
import org.netkernel.layer0.nkf.INKFRequestContext;
import org.netkernel.layer0.nkf.INKFRequestReadOnly;
import org.netkernel.layer0.nkf.INKFResponseReadOnly;
import org.netkernel.layer0.nkf.NKFException;
import org.netkernel.layer0.nkf.impl.NKFResponseReadOnlyImpl;
import org.netkernel.layer0.util.PairList;
import org.netkernel.layer0.util.RequestScopeClassLoader;
import org.netkernel.request.IRequestScopeLevel;
import org.netkernel.request.impl.ExpiryFactory;
import org.netkernel.request.impl.MetaImpl;
import org.netkernel.request.impl.RequestFactory;
import org.netkernel.request.impl.RequestScopeLevelImpl;
import org.netkernel.urii.ISpace;

import java.util.List;


/** A sub request with arguments
 * @author Tony Butterfield
 *
 */
public class SubRequest extends VariableImpl implements IValueVariable
{
	private final String mRequestId;
	private final int mVerb;
	private final List mArguments;
	private final String mRepName;
	private final ISpace mDPMLContext;
	
	public SubRequest(String aName, IScope aScope, ILocation aLocation, String aRequestId, int aVerb, List aArguments, String aRepName, ISpace aDPMLContext)
	{	super(aName,aScope,aLocation);
		mRequestId=aRequestId;
		mVerb=aVerb;
		mArguments=aArguments;
		mRepName=aRepName;
		mDPMLContext=aDPMLContext;
	}
	
	public IValue getValue(INKFRequestContext aContext, IRequestScopeLevel aSubRequestContext, Class aRepresentationClass) throws NKFException
	{
		INKFRequest req = aContext.createRequest(mRequestId);
		req.setVerb(mVerb);
		if (mRepName!=null)
		{
			RequestScopeClassLoader cl = new RequestScopeClassLoader(aSubRequestContext);
			try
			{	Class repClass = cl.loadClass(mRepName);
				req.setRepresentationClass(repClass);
			}
			catch (Exception e)
			{	throw aContext.createFormattedException("EX_DPML_RUNTIME", "MSG_REP_NOT_FOUND", DPMLRuntime.formatLocationMessage(this,aContext), e, mRepName);
			}
		}
		else
		{	req.setRepresentationClass(aRepresentationClass);
		}
		
		DPMLSubrequestContext ctx = null;
		
		/*
		IRequestScopeLevel scopeToUseForRequests;
		if (aSubRequestContext.getSpace() instanceof SequenceAssignmentsSpace)
		{	scopeToUseForRequests=aSubRequestContext.getParent();
		}
		else
		{	scopeToUseForRequests=aSubRequestContext;
		}
		*/
		
		boolean debug= DPMLRuntime.shouldLog();
		PairList argDebug=null;
		if (debug)
		{	argDebug=new PairList(mArguments.size());
		}

		
		for (int i=0; i<mArguments.size(); i++)
		{	IArgument arg=(IArgument)mArguments.get(i);
			String name=arg.getName();
			if (name.equals(IArgument.ARG_PRIMARY))
			{
				IValueVariable var=(IValueVariable) DPMLRuntime.processArgument(arg,aContext,aSubRequestContext,true);
				IValue value=var.getValue(aContext,aSubRequestContext,Object.class);
				req.addPrimaryArgumentFromResponse(DPMLRuntime.createResponseWithMetaData(value,aContext));
			}
			else
			{
				IVariable var=arg.getVariable();
				
				var= DPMLRuntime.processArgument(arg,aContext,aSubRequestContext,true);
				String identifier;
				if (var instanceof IIdentifierVariable)
				{	identifier = ((IIdentifierVariable)var).getIdentifier(aContext,aSubRequestContext);
					if (((IIdentifierVariable)var).isAbsolute())
					{	req.addArgumentAsAbsolute( name,  identifier);
					}
					else
					{	req.addArgument( name,  identifier);
					}
				}
				else
				{	
					if (var.getScope()!=null)
					{	identifier= DPMLScriptContext.CLOSURE_SCHEME+var.getId();
						req.addArgument( name,  identifier);
					}
					else
					{	if (ctx==null)
						{	ctx=new DPMLSubrequestContext(aSubRequestContext,mDPMLContext,aContext.getCWU());
							try
							{	ctx.onCommissionSpace(aContext.getKernelContext().getKernel());
							} catch (Exception e) {}
						}
						ctx.putArgument(name,var);
						identifier= DPMLSubrequestContext.ARG_SCHEME_ESC+name;
						req.addArgument( name,  identifier);
					}
				}
				
				if (debug)
				{	if (var.getName()!=null)
					{	argDebug.put(name,"this:"+var.getName());
					}
					else
					{	argDebug.put(name,identifier);
					}
				}
			}
		}
		
		if (DPMLRuntime.shouldLog())
		{	
			StringBuffer sb=new StringBuffer(64);
			final int s=argDebug.size();
			for (int i=0; i<s; i++)
			{	sb.append(argDebug.getValue1(i));
				sb.append('=');
				sb.append(argDebug.getValue2(i));
				if (i<(s-1))
				{	sb.append(", ");
				}
				else
				{	break;
				}
			}
			String argString = sb.toString();
			Object[] args = new Object[] {  getLocation().getSourceIdentity(),RequestFactory.verbString(mVerb),mRequestId, argString};
			String msg=aContext.formatMessage("MSG_LOG_REQ",args);
			ILogger logger=aContext.getKernelContext().getKernel().getLogger();
			logger.logRaw(ILogger.DEBUG, DPMLRuntime.getInstance(),msg);
		}
		
			
		if (ctx!=null)
		{	aSubRequestContext = RequestScopeLevelImpl.createDurableScopeLevel(ctx,aSubRequestContext);
		}
		
		req.setRequestScope(aSubRequestContext);

		try
		{	INKFResponseReadOnly resp=aContext.issueRequestForResponse(req);
			return new ValueImpl(resp.getRepresentation(),((NKFResponseReadOnlyImpl)resp).getKernelResponse().getMeta());
		}
		catch (NKFException e)
		{	if (mVerb==INKFRequestReadOnly.VERB_EXISTS)
			{	return new ValueImpl(Boolean.FALSE,new MetaImpl(ExpiryFactory.getDependentExpiry()));
			}
			else
			{	throw e;
			}
		}
		
	}
	
}
