package org.netkernel.lang.dpml.endpoint;

import org.netkernel.layer0.nkf.INKFRequestContext;
import org.netkernel.layer0.nkf.INKFResponse;
import org.netkernel.layer0.nkf.INKFResponseReadOnly;
import org.netkernel.layer0.util.FastSchematron;
import org.netkernel.layer0.util.PairList;
import org.netkernel.layer0.util.RequestBuilder;
import org.netkernel.layer0.util.RequestScopeClassLoader;
import org.netkernel.layer0.util.XMLReadable;
import org.netkernel.module.standard.endpoint.StandardAccessorImpl;
import org.netkernel.xml.util.XMLUtils;
import org.w3c.dom.*;

public class ResponseAccessor extends StandardAccessorImpl
{
	private static final PairList VALIDATION = new PairList(4);
	static
	{	VALIDATION.put("/*[name()!='config']", "MSG_MR_BAD_ROOT");
		VALIDATION.put("/config/*[name()!='header' and name()!='expiry']", "MSG_MR_BAD_TAG1");
		VALIDATION.put("/config[count(expiry)>1]", "MSG_MR_MANY_EXPIRY");
		VALIDATION.put("/config/expiry[not(@method)]", "MSG_MR_EXPIRY_NO_METHOD");
		VALIDATION.put("/config/header[not(@name)]", "MSG_MR_HEADER_NO_NAME");
		VALIDATION.put("/config/header[count(*)>1]", "MSG_MR_HEADER_MULTILITERAL");
		VALIDATION.put("/config/header/*[name()!='literal']", "MSG_MR_HEADER_MULTILITERAL");
	}	
	
	//This header indicates to dpml that response has been modified so it's expiry should be used rather than defaulting to DEPENDENT
	public static final String HEADER_EXPIRY="dpml:expiry";
	public static final String HEADER_EXPIRY_PERIOD="dpml:expiry-time";
	
	public ResponseAccessor()
	{	declareThreadSafe();
		declareUnhandledExceptionsExpired(false);
	}
	
	public void onSource(INKFRequestContext aContext) throws Exception
	{
		INKFResponseReadOnly responseIn=aContext.sourceForResponse("arg:operand");
		Document configDoc=aContext.source("arg:config",Document.class);
		
		PairList validation=FastSchematron.validate(configDoc.getDocumentElement(), VALIDATION);
		if (validation.size()>0)
		{	String m = FastSchematron.formatValidationError(validation,aContext);
			throw aContext.createFormattedException("EX_MR_SYNTAX", "MSG_RAW", null, null, m);
		}
		XMLReadable r=new XMLReadable(configDoc.getDocumentElement());
		
		INKFResponse response = aContext.createResponseFrom(responseIn);
		
		//add all headers
		for (Node headerNode: r.getNodes("header"))
		{	String name=((Element)headerNode).getAttribute("name");
			Object value;
			Element literalNode=XMLUtils.getFirstChildElement(headerNode);
			if (literalNode==null)
			{	value=XMLUtils.getText(headerNode);
			}
			else if (XMLUtils.getNextSiblingElement(literalNode)==null)
			{	ClassLoader cl=new RequestScopeClassLoader(aContext.getKernelContext().getRequestScope());
				value=RequestBuilder.parseLiteral(literalNode, cl, aContext)[0];
			}
			else
			{	throw aContext.createFormattedException("EX_MR_SYNTAX", "MSG_MR_HEADER_MULTILITERAL", null, null);
			}
			response.setHeader(name, value);
		}
		
		String expiry=r.getText("expiry/@method");
		
		if (expiry.length()>0)
		{	String periodString=r.getText("expiry").trim();
			long period=-1;
			if (periodString.length()>0)
			{	try
				{	period=Long.parseLong(periodString);
				}
				catch (NumberFormatException e)
				{ }
				if (period<0)
				{	throw aContext.createFormattedException("EX_MR_SYNTAX", "MSG_MR_EXPIRY_PERIOD", null, null, periodString);
				}
			}
			int exp;
			if (expiry.equals("ALWAYS"))
			{	exp=INKFResponse.EXPIRY_ALWAYS;
			}
			else if (expiry.equals("NEVER"))
			{	exp=INKFResponse.EXPIRY_NEVER;
			}
			else if (expiry.equals("DEPENDENT"))
			{	exp=INKFResponse.EXPIRY_DEPENDENT;
			}
			else if (expiry.startsWith("CONSTANT"))
			{	exp=INKFResponse.EXPIRY_CONSTANT;
			}
			else if (expiry.startsWith("MIN_CONSTANT_DEPENDENT"))
			{	exp=INKFResponse.EXPIRY_MIN_CONSTANT_DEPENDENT;
			}
			else if (expiry.startsWith("MAX_CONSTANT_DEPENDENT"))
			{	exp=INKFResponse.EXPIRY_MAX_CONSTANT_DEPENDENT;
			}
			else
			{	throw aContext.createFormattedException("EX_MR_SYNTAX", "MSG_MR_UNSUPPORTED_EXPIRY", null, null, expiry);
			}
			response.setHeader(HEADER_EXPIRY, exp);
			response.setHeader(HEADER_EXPIRY_PERIOD, period);
		}
		
	}		
}
